# A basic auto-responder bot that replies to specific messages when bot mode is enabled.
import logging
import threading
import time
from typing import Any, Dict

import contact.ui.default_config as config
from contact.utilities.singleton import app_state, interface_state, ui_state
from contact.message_handlers.tx_handler import send_message

BOT_RESPONSE_DELAY_SECONDS = 2.3

def _get_bot_catch_words() -> set[str]:
    """Return normalized bot trigger words from app settings."""
    raw_words = getattr(config, "ping_bot_catch_words", "ping; test")
    words = {
        word.strip().casefold()
        for word in raw_words.replace(";", ",").split(",")
        if word.strip()
    }
    return words or {"ping"}

def is_bot_message(message: str) -> bool:
    """Return True when the incoming message should trigger an automatic response."""
    return message.strip().casefold() in _get_bot_catch_words()

def bot_respond(packet: Dict[str, Any], message: str, send_channel: int) -> bool:
    """Send a basic response when bot mode is enabled."""
    if not ui_state.bot_mode_enabled:
        return False

    if not is_bot_message(message):
        """ Only respond to specific messages. """
        return False

    from_node = packet.get("from")
    if from_node is None:
        return False
    if from_node == interface_state.myNodeNum:
        return False
    snr = packet.get('rxSnr', -128)
    rssi = packet.get('rxRssi', -128)
    replyIDset = packet.get('replyId', False)
    hop_start = packet.get('hopStart', 0)
    hop_limit = packet.get('hopLimit', 0)
    transport_type = packet.get('transportMechanism', None)
    hops = hop_start - hop_limit
    
    details = []
    if snr != -128:
        details.append(f"SNR: {snr}")
    if rssi != -128:
        details.append(f"RSSI: {rssi}")
    if hops != 0:
        details.append(f"Hops: {hops}")
    if replyIDset:
        details.append(f"Relay: {replyIDset}")
    transport_text = str(transport_type).upper() if transport_type is not None else ""
    for transport_name in ("UDP", "MQTT"):
        if transport_name in transport_text:
            details.append(f"Via: {transport_name}")

    response_data_string = getattr(config, "ping_bot_response_word", "Pong!")
    if details:
        response_data_string += f" {', '.join(details)}"

    def send_response_delayed() -> None:
        try:
            time.sleep(BOT_RESPONSE_DELAY_SECONDS)

            with app_state.lock:
                if not ui_state.bot_mode_enabled:
                    return

                send_message(response_data_string,channel=send_channel)

            # Import locally to avoid circular import at module import time.
            from contact.ui.contact_ui import request_ui_redraw

            request_ui_redraw(channels=True, messages=True, scroll_messages_to_bottom=True)
            logging.info("Bot response sent to %s on channel index %s", from_node, send_channel)
        except Exception:
            logging.exception("Bot response send failed for destination %s", from_node)

    threading.Thread(target=send_response_delayed, name="bot-response", daemon=True).start()

    return True



#!/usr/bin/env python3

"""
Contact - A Console UI for Meshtastic by http://github.com/pdxlocations
Powered by Meshtastic.org

Meshtastic® is a registered trademark of Meshtastic LLC.
Meshtastic software components are released under various licenses—see GitHub for details.
No warranty is provided. Use at your own risk.
"""

# Standard library
import contextlib
import curses
import io
import logging
import os
import subprocess
import sys
import threading
import traceback
from typing import Optional

# Third-party
from pubsub import pub

# Local application
import contact.ui.default_config as config
from contact.message_handlers.rx_handler import on_receive
from contact.settings import set_region
from contact.ui.colors import setup_colors
from contact.ui.contact_ui import main_ui
from contact.ui.splash import draw_splash
from contact.utilities.arg_parser import setup_parser
from contact.utilities.db_handler import init_nodedb, load_messages_from_db
from contact.utilities.demo_data import build_demo_interface, configure_demo_database, seed_demo_messages
from contact.utilities.input_handlers import get_list_input
from contact.utilities.i18n import t
from contact.ui.dialog import dialog
from contact.utilities.interfaces import initialize_interface, reconnect_interface
from contact.utilities.utils import get_channels, get_nodeNum, get_node_list
from contact.utilities.singleton import ui_state, interface_state, app_state

# ------------------------------------------------------------------------------
# Environment & Logging Setup
# ------------------------------------------------------------------------------

os.environ["NCURSES_NO_UTF8_ACS"] = "1"
os.environ["LANG"] = "C.UTF-8"
os.environ.setdefault("TERM", "xterm-256color")
if os.environ.get("COLORTERM") == "gnome-terminal":
    os.environ["TERM"] = "xterm-256color"

logging.basicConfig(
    filename=config.log_file_path, level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

app_state.lock = threading.Lock()

DEFAULT_CLOSE_TIMEOUT_SECONDS = 5.0


# ------------------------------------------------------------------------------
# Main Program Logic
# ------------------------------------------------------------------------------
def prompt_region_if_unset(args: object, stdscr: Optional[curses.window] = None) -> None:
    """Prompt user to set region if it is unset."""
    confirmation = get_list_input("Your region is UNSET. Set it now?", "Yes", ["Yes", "No"])
    if confirmation == "Yes":
        set_region(interface_state.interface)
        close_interface(interface_state.interface)
        if stdscr is not None:
            draw_splash(stdscr)
        interface_state.interface = reconnect_interface(args)


def close_interface(interface: object, timeout_seconds: float = DEFAULT_CLOSE_TIMEOUT_SECONDS) -> bool:
    if interface is None:
        return True

    close_errors = []

    def _close_target() -> None:
        try:
            interface.close()
        except BaseException as error:  # Keep shutdown resilient even for KeyboardInterrupt/SystemExit from libraries.
            close_errors.append(error)

    close_thread = threading.Thread(target=_close_target, name="meshtastic-interface-close", daemon=True)
    close_thread.start()
    close_thread.join(timeout_seconds)

    if close_thread.is_alive():
        logging.warning("Timed out closing interface after %.1fs; continuing shutdown", timeout_seconds)
        return False

    if not close_errors:
        return True

    error = close_errors[0]
    if isinstance(error, KeyboardInterrupt):
        logging.info("Interrupted while closing interface; continuing shutdown")
        return True

    logging.warning("Ignoring error while closing interface: %r", error)
    return True


def interface_is_ready(interface: object) -> bool:
    try:
        return getattr(interface, "localNode", None) is not None and interface.localNode.localConfig is not None
    except Exception:
        return False


def initialize_runtime_interface_with_retry(stdscr: curses.window, args: object):
    while True:
        interface = initialize_runtime_interface(args)
        if getattr(args, "demo_screenshot", False) or interface_is_ready(interface):
            return interface

        choice = get_list_input(
            t("ui.prompt.node_not_found", default="No node found. Retry connection?"),
            "Retry",
            ["Retry", "Close"],
            mandatory=True,
        )
        close_interface(interface)
        if choice == "Close":
            return None

        draw_splash(stdscr)


def initialize_globals(seed_demo: bool = False) -> None:
    """Initializes interface and shared globals."""

    ui_state.channel_list = []
    ui_state.all_messages = {}
    ui_state.notifications = []
    ui_state.packet_buffer = []
    ui_state.node_list = []
    ui_state.selected_channel = 0
    ui_state.selected_message = 0
    ui_state.selected_node = 0
    ui_state.start_index = [0, 0, 0]
    interface_state.myNodeNum = get_nodeNum()
    ui_state.channel_list = get_channels()
    ui_state.node_list = get_node_list()
    ui_state.single_pane_mode = config.single_pane_mode.lower() == "true"
    pub.subscribe(on_receive, "meshtastic.receive")

    init_nodedb()
    if seed_demo:
        seed_demo_messages()
    load_messages_from_db()


def initialize_runtime_interface(args: object):
    if getattr(args, "demo_screenshot", False):
        configure_demo_database()
        return build_demo_interface()
    return initialize_interface(args)


def main(stdscr: curses.window) -> None:
    """Main entry point for the curses UI."""

    output_capture = io.StringIO()
    try:
        setup_colors()
        ensure_min_rows(stdscr)
        draw_splash(stdscr)

        args = setup_parser().parse_args()

        if getattr(args, "settings", False):
            subprocess.run([sys.executable, "-m", "contact.settings"], check=True)
            return

        logging.info("Initializing interface...")
        with app_state.lock:
            interface_state.interface = initialize_runtime_interface_with_retry(stdscr, args)
            if interface_state.interface is None:
                return

            if not getattr(args, "demo_screenshot", False) and interface_state.interface.localNode.localConfig.lora.region == 0:
                prompt_region_if_unset(args, stdscr)

            initialize_globals(seed_demo=getattr(args, "demo_screenshot", False))
            logging.info("Starting main UI")

            stdscr.clear()
            stdscr.refresh()

        try:
            with contextlib.redirect_stdout(output_capture), contextlib.redirect_stderr(output_capture):
                main_ui(stdscr)
        except Exception:
            console_output = output_capture.getvalue()
            logging.error("Uncaught exception inside main_ui")
            logging.error("Traceback:\n%s", traceback.format_exc())
            logging.error("Console output:\n%s", console_output)
            return

    except Exception:
        raise


def ensure_min_rows(stdscr: curses.window, min_rows: int = 11) -> None:
    while True:
        rows, _ = stdscr.getmaxyx()
        if rows >= min_rows:
            return
        dialog(
            t("ui.dialog.resize_title", default="Resize Terminal"),
            t(
                "ui.dialog.resize_body",
                default="Please resize the terminal to at least {rows} rows.",
                rows=min_rows,
            ),
        )
        curses.update_lines_cols()
        stdscr.clear()
        stdscr.refresh()


def start() -> None:
    """Entry point for the application."""

    if "--help" in sys.argv or "-h" in sys.argv:
        setup_parser().print_help()
        sys.exit(0)

    interrupted = False
    fatal_error = None

    try:
        curses.wrapper(main)
    except KeyboardInterrupt:
        interrupted = True
        logging.info("User exited with Ctrl+C")
    except Exception as e:
        fatal_error = e
        logging.critical("Fatal error", exc_info=True)
        try:
            curses.endwin()
        except Exception:
            pass
    finally:
        close_interface(interface_state.interface)

    if fatal_error is not None:
        print("Fatal error:", fatal_error)
        traceback.print_exc()
        sys.exit(1)

    if interrupted:
        sys.exit(0)


if __name__ == "__main__":
    start()



import time

from typing import Any, Dict

import google.protobuf.json_format
from meshtastic import BROADCAST_NUM
from meshtastic.protobuf import mesh_pb2, portnums_pb2

from contact.utilities.db_handler import (
    save_message_to_db,
    update_ack_nak,
    get_name_from_database,
    is_chat_archived,
    update_node_info_in_db,
)
import contact.ui.default_config as config

from contact.utilities.singleton import ui_state, interface_state, app_state

from contact.utilities.utils import add_new_message

ack_naks: Dict[str, Dict[str, Any]] = {}  # requestId -> {channel, messageIndex, timestamp}


# Note "onAckNak" has special meaning to the API, thus the nonstandard naming convention
# See https://github.com/meshtastic/python/blob/master/meshtastic/mesh_interface.py#L462
def onAckNak(packet: Dict[str, Any]) -> None:
    """
    Handles incoming ACK/NAK response packets.
    """
    from contact.ui.contact_ui import request_ui_redraw

    with app_state.lock:
        request = packet["decoded"]["requestId"]
        if request not in ack_naks:
            return

        acknak = ack_naks.pop(request)
        message = ui_state.all_messages[acknak["channel"]][acknak["messageIndex"]][1]

        confirm_string = " "
        ack_type = None
        if packet["decoded"]["routing"]["errorReason"] == "NONE":
            if packet["from"] == interface_state.myNodeNum:  # Ack "from" ourself means implicit ACK
                confirm_string = config.ack_implicit_str
                ack_type = "Implicit"
            else:
                confirm_string = config.ack_str
                ack_type = "Ack"
        else:
            confirm_string = config.nak_str
            ack_type = "Nak"

        ui_state.all_messages[acknak["channel"]][acknak["messageIndex"]] = (
            time.strftime("[%H:%M:%S] ") + config.sent_message_prefix + confirm_string + ": ",
            message,
        )

        update_ack_nak(acknak["channel"], acknak["timestamp"], message, ack_type)

        channel_number = ui_state.channel_list.index(acknak["channel"])
        if ui_state.channel_list[channel_number] == ui_state.channel_list[ui_state.selected_channel]:
            request_ui_redraw(messages=True)


def on_response_traceroute(packet: Dict[str, Any]) -> None:
    """
    Handle traceroute response packets and render the route visually in the UI.
    """
    from contact.ui.contact_ui import add_notification, request_ui_redraw

    with app_state.lock:
        refresh_channels = False
        refresh_messages = False

        UNK_SNR = -128  # Value representing unknown SNR

        route_discovery = mesh_pb2.RouteDiscovery()
        route_discovery.ParseFromString(packet["decoded"]["payload"])
        msg_dict = google.protobuf.json_format.MessageToDict(route_discovery)

        msg_str = "Traceroute to:\n"

        route_str = (
            get_name_from_database(packet["to"], "short") or f"{packet['to']:08x}"
        )  # Start with destination of response

        lenTowards = 0 if "route" not in msg_dict else len(msg_dict["route"])
        snrTowardsValid = "snrTowards" in msg_dict and len(msg_dict["snrTowards"]) == lenTowards + 1
        if lenTowards > 0:
            for idx, node_num in enumerate(msg_dict["route"]):
                route_str += (
                    " --> "
                    + (get_name_from_database(node_num, "short") or f"{node_num:08x}")
                    + " ("
                    + (
                        str(msg_dict["snrTowards"][idx] / 4)
                        if snrTowardsValid and msg_dict["snrTowards"][idx] != UNK_SNR
                        else "?"
                    )
                    + "dB)"
                )

        route_str += (
            " --> "
            + (get_name_from_database(packet["from"], "short") or f"{packet['from']:08x}")
            + " ("
            + (str(msg_dict["snrTowards"][-1] / 4) if snrTowardsValid and msg_dict["snrTowards"][-1] != UNK_SNR else "?")
            + "dB)"
        )

        msg_str += route_str + "\n"

        lenBack = 0 if "routeBack" not in msg_dict else len(msg_dict["routeBack"])
        backValid = "hopStart" in packet and "snrBack" in msg_dict and len(msg_dict["snrBack"]) == lenBack + 1
        if backValid:
            msg_str += "Back:\n"
            route_str = get_name_from_database(packet["from"], "short") or f"{packet['from']:08x}"

            if lenBack > 0:
                for idx, node_num in enumerate(msg_dict["routeBack"]):
                    route_str += (
                        " --> "
                        + (get_name_from_database(node_num, "short") or f"{node_num:08x}")
                        + " ("
                        + (str(msg_dict["snrBack"][idx] / 4) if msg_dict["snrBack"][idx] != UNK_SNR else "?")
                        + "dB)"
                    )

            route_str += (
                " --> "
                + (get_name_from_database(packet["to"], "short") or f"{packet['to']:08x}")
                + " ("
                + (str(msg_dict["snrBack"][-1] / 4) if msg_dict["snrBack"][-1] != UNK_SNR else "?")
                + "dB)"
            )

            msg_str += route_str + "\n"

        if packet["from"] not in ui_state.channel_list:
            ui_state.channel_list.append(packet["from"])
            refresh_channels = True

        if is_chat_archived(packet["from"]):
            update_node_info_in_db(packet["from"], chat_archived=False)

        channel_number = ui_state.channel_list.index(packet["from"])
        channel_id = ui_state.channel_list[channel_number]

        if channel_id == ui_state.channel_list[ui_state.selected_channel]:
            refresh_messages = True
        else:
            add_notification(channel_number)
            refresh_channels = True

        message_from_string = get_name_from_database(packet["from"], type="short") + ":\n"

        add_new_message(channel_id, f"{config.message_prefix} {message_from_string}", msg_str)

        if refresh_channels:
            request_ui_redraw(channels=True)
        if refresh_messages:
            request_ui_redraw(messages=True, scroll_messages_to_bottom=True)

        save_message_to_db(channel_id, packet["from"], msg_str)

def send_message(message: str, destination: int = BROADCAST_NUM, channel: int = 0) -> None:
    """
    Sends a chat message using the selected channel.
    """
    myid = interface_state.myNodeNum
    send_on_channel = 0
    channel_id = ui_state.channel_list[channel]
    if isinstance(channel_id, int):
        send_on_channel = 0
        destination = channel_id
    elif isinstance(channel_id, str):
        send_on_channel = channel

    sent_message_data = interface_state.interface.sendText(
        text=message,
        destinationId=destination,
        wantAck=True,
        wantResponse=False,
        onResponse=onAckNak,
        channelIndex=send_on_channel,
    )

    add_new_message(channel_id, config.sent_message_prefix + config.ack_unknown_str + ": ", message)

    timestamp = save_message_to_db(channel_id, myid, message)

    ack_naks[sent_message_data.id] = {
        "channel": channel_id,
        "messageIndex": len(ui_state.all_messages[channel_id]) - 1,
        "timestamp": timestamp,
    }


def send_traceroute() -> None:
    """
    Sends a RouteDiscovery protobuf to the selected node.
    """

    channel_id = ui_state.node_list[ui_state.selected_node]
    add_new_message(channel_id, f"{config.message_prefix} Sent Traceroute", "")

    r = mesh_pb2.RouteDiscovery()
    interface_state.interface.sendData(
        r,
        destinationId=channel_id,
        portNum=portnums_pb2.PortNum.TRACEROUTE_APP,
        wantResponse=True,
        onResponse=on_response_traceroute,
        channelIndex=0,
        hopLimit=3,
    )
